#include "SceneGraph.h"
#include "KdTree.h"
#include "RssPreprocessor.h"
#include "X3dExporter.h"
#include "Environment.h"
#include "MutualVisibility.h"
#include "Polygon3.h"
#include "ViewCell.h"
#include "VssRay.h"
#include "RssTree.h"
#include "ViewCellsManager.h"
#include "RenderSimulator.h"

static bool useViewSpaceBox = true;
static bool use2dSampling = false;
static bool fromBoxVisibility = false;

static bool exportPvs = false;
static bool exportRssTree = false;

static int numExportRays = 5000;
// static int numExportRays = 0;

static bool useViewcells = false;

RssPreprocessor::RssPreprocessor():
  mPass(0),
  mVssRays()
{
  // this should increase coherence of the samples
  environment->GetIntValue("RssPreprocessor.samplesPerPass", mSamplesPerPass);
  environment->GetIntValue("RssPreprocessor.initialSamples", mInitialSamples);
  environment->GetIntValue("RssPreprocessor.vssSamples", mRssSamples);
  environment->GetIntValue("RssPreprocessor.vssSamplesPerPass", mRssSamplesPerPass);
  environment->GetBoolValue("RssPreprocessor.useImportanceSampling", mUseImportanceSampling);
  environment->GetIntValue("BspTree.Construction.samples", mBspConstructionSamples);
	
  mStats.open("stats.log");
}

RssPreprocessor::~RssPreprocessor()
{
  CLEAR_CONTAINER(mVssRays);
}

void
RssPreprocessor::SetupRay(Ray &ray, 
						  const Vector3 &point, 
						  const Vector3 &direction
						  )
{
  ray.intersections.clear();
  // do not store anything else then intersections at the ray
  ray.Init(point, direction, Ray::LOCAL_RAY);
}

int
RssPreprocessor::CastRay(
						 Vector3 &viewPoint,
						 Vector3 &direction,
						 VssRayContainer &vssRays
						 )
{
  int hits = 0;
  static Ray ray;
  AxisAlignedBox3 box = mKdTree->GetBox();

  AxisAlignedBox3 sbox = box;
  sbox.Enlarge(Vector3(-Limits::Small));
  if (!sbox.IsInside(viewPoint))
	return 0;
	
  SetupRay(ray, viewPoint, direction);
  // cast ray to KD tree to find intersection with other objects
  Intersectable *objectA, *objectB;
  Vector3 pointA, pointB;
  float bsize = Magnitude(box.Size());
  
  if (mKdTree->CastRay(ray)) {
	objectA = ray.intersections[0].mObject;
	pointA = ray.Extrap(ray.intersections[0].mT);
  } else {
	objectA = NULL;
	// compute intersection with the scene bounding box
	float tmin, tmax;
	box.ComputeMinMaxT(ray, &tmin, &tmax);
	if (tmax > bsize) {
	  //cerr<<"Warning: tmax > box size tmax="<<tmax<<" tmin="<<tmin<<" size="<<bsize<<endl;
	  //cerr<<"ray"<<ray<<endl;
	}
	pointA = ray.Extrap(tmax);
  }

  bool detectEmptyViewSpace = true;
	
  if (detectEmptyViewSpace) {
	SetupRay(ray, pointA, -direction);
  } else
	SetupRay(ray, viewPoint, -direction);
	
	
  if (mKdTree->CastRay(ray)) {
	objectB = ray.intersections[0].mObject;
	pointB = ray.Extrap(ray.intersections[0].mT);
  } else {
	objectB = NULL;
	float tmin, tmax;
	box.ComputeMinMaxT(ray, &tmin, &tmax);
	if (tmax > bsize) {
	  //cerr<<"Warning: tmax > box size tmax="<<tmax<<" tmin="<<tmin<<" size="<<bsize<<endl;
	  //cerr<<"ray"<<ray<<endl;
	}
	pointB = ray.Extrap(tmax);
  }

  VssRay *vssRay  = NULL;

  bool validSample = true;
  if (detectEmptyViewSpace) {
	// check if the viewpoint lies on the line segment AB
	if (Distance(pointA, pointB) <
		Distance(viewPoint, pointA) + Distance(viewPoint, pointB) - Limits::Small) {
	  validSample = false;
	}
  }
	
  if (validSample) {
	if (objectA) {
	  vssRay = new VssRay(pointB,
						  pointA,
						  objectB,
						  objectA);
	  vssRays.push_back(vssRay);
	  hits ++;
	}
	
	if (objectB) {
	  vssRay = new VssRay(pointA,
						  pointB,
						  objectA,
						  objectB);
	  vssRays.push_back(vssRay);
	  hits ++;
	}
  }
	
  return hits;
}


Vector3
RssPreprocessor::GetViewpoint(AxisAlignedBox3 *viewSpaceBox)
{
  AxisAlignedBox3 box;
	
  if (viewSpaceBox)
	box =*viewSpaceBox;
  else 
	box = mKdTree->GetBox();
	
  // shrink the box in the y direction
  return box.GetRandomPoint();
}

Vector3
RssPreprocessor::GetDirection(const Vector3 &viewpoint,
							  AxisAlignedBox3 *viewSpaceBox
							  )
{
  Vector3 point;
  if (!use2dSampling) {
	Vector3 normal;
	int i = RandomValue(0, mObjects.size()-1);
	Intersectable *object = mObjects[i];
	object->GetRandomSurfacePoint(point, normal);
  } else {
	AxisAlignedBox3 box;
		
	if (viewSpaceBox)
	  box =*viewSpaceBox;
	else 
	  box = mKdTree->GetBox();
		
	point = box.GetRandomPoint();
	point.y = viewpoint.y;
  }
	
  return point - viewpoint;
}

int
RssPreprocessor::GenerateImportanceRays(RssTree *rssTree,
										const int desiredSamples,
										SimpleRayContainer &rays
										)
{
  int num;


  
  float avgPvsSize;
  float avgRayContribution;
  float avgPvsEntropy;
  float avgRayLengthEntropy;
  float avgImportance;
  float avgRays;
  rssTree->GetTreeStatistics(
							 avgPvsSize,
							 avgRays,
							 avgRayContribution,
							 avgPvsEntropy,
							 avgRayLengthEntropy,
							 avgImportance);
  
  cout<<
	"#AVG_PVS_SIZE\n"<<avgPvsSize<<endl<<
	"#AVG_RAYS\n"<<avgRays<<endl<<
	"#AVG_RAY_CONTRIB\n"<<avgRayContribution<<endl<<
	"#AVG_PVS_ENTROPY\n"<<avgPvsEntropy<<endl<<
	"#AVG_RAY_LENGTH_ENTROPY\n"<<avgRayLengthEntropy<<endl<<
	"#AVG_IMPORTANCE\n"<<avgImportance<<endl;
  
  if (0) {
	float p = desiredSamples/(float)(avgRayContribution*rssTree->stat.Leaves());
	num = rssTree->GenerateRays(p, rays);
  } else {
	int leaves = rssTree->stat.Leaves()/1;
	num = rssTree->GenerateRays(desiredSamples, leaves, rays);
  }
	
  cout<<"Generated "<<num<<" rays."<<endl;
	
  return num;
}


bool
RssPreprocessor::ExportRays(const char *filename,
							const VssRayContainer &vssRays,
							const int number
							)
{
  cout<<"Exporting vss rays..."<<endl<<flush;
	
  float prob = number/(float)vssRays.size();


  Exporter *exporter = NULL;
  exporter = Exporter::GetExporter(filename);
  //	exporter->SetWireframe();
  //	exporter->ExportKdTree(*mKdTree);
  exporter->SetFilled();
  exporter->ExportScene(mSceneGraph->mRoot);
  exporter->SetWireframe();

  if (mViewSpaceBox) {
	exporter->SetForcedMaterial(RgbColor(1,0,1));
	exporter->ExportBox(*mViewSpaceBox);
	exporter->ResetForcedMaterial();
  }
	
  VssRayContainer rays;	for (int i=0; i < vssRays.size(); i++)
	if (RandomValue(0,1) < prob)
	  rays.push_back(vssRays[i]);

  exporter->ExportRays(rays, RgbColor(1, 0, 0));
	
  delete exporter;

  cout<<"done."<<endl<<flush;

  return true;
}


bool
RssPreprocessor::ExportRssTree(char *filename,
							   RssTree *tree,
							   const Vector3 &dir
							   )
{
  Exporter *exporter = Exporter::GetExporter(filename);
  exporter->SetFilled();
  exporter->ExportScene(mSceneGraph->mRoot);
  //  exporter->SetWireframe();
  bool result = exporter->ExportRssTree2( *tree, dir );
  delete exporter;
  return result;
}

bool
RssPreprocessor::ExportRssTreeLeaf(char *filename,
								   RssTree *tree,
								   RssTreeLeaf *leaf)
{
  Exporter *exporter = NULL;
  exporter = Exporter::GetExporter(filename);
  exporter->SetWireframe();
  exporter->ExportKdTree(*mKdTree);
	
  if (mViewSpaceBox) {
	exporter->SetForcedMaterial(RgbColor(1,0,0));
	exporter->ExportBox(*mViewSpaceBox);
	exporter->ResetForcedMaterial();
  }
	
  exporter->SetForcedMaterial(RgbColor(0,0,1));
  exporter->ExportBox(tree->GetBBox(leaf));
  exporter->ResetForcedMaterial();
	
  VssRayContainer rays[4];
  for (int i=0; i < leaf->rays.size(); i++) {
	int k = leaf->rays[i].GetRayClass();
	rays[k].push_back(leaf->rays[i].mRay);
  }
	
  // SOURCE RAY
  exporter->ExportRays(rays[0], RgbColor(1, 0, 0));
  // TERMINATION RAY
  exporter->ExportRays(rays[1], RgbColor(1, 1, 1));
  // PASSING_RAY
  exporter->ExportRays(rays[2], RgbColor(1, 1, 0));
  // CONTAINED_RAY
  exporter->ExportRays(rays[3], RgbColor(0, 0, 1));

  delete exporter;
  return true;
}

void
RssPreprocessor::ExportRssTreeLeaves(RssTree *tree, const int number)
{
  vector<RssTreeLeaf *> leaves;
  tree->CollectLeaves(leaves);

  int num = 0;
  int i;
  float p = number / (float)leaves.size();
  for (i=0; i < leaves.size(); i++) {
	if (RandomValue(0,1) < p) {
	  char filename[64];
	  sprintf(filename, "rss-leaf-%04d.x3d", num);
	  ExportRssTreeLeaf(filename, tree, leaves[i]);
	  num++;
	}
	if (num >= number)
	  break;
  }
}


float
RssPreprocessor::GetAvgPvsSize(RssTree *tree,
							   const vector<AxisAlignedBox3> &viewcells
							   )
{
  vector<AxisAlignedBox3>::const_iterator it, it_end = viewcells.end();

  int sum = 0;
  for (it = viewcells.begin(); it != it_end; ++ it)
	sum += tree->GetPvsSize(*it);
	
  return sum/(float)viewcells.size();
}


void
RssPreprocessor::ExportPvs(char *filename,
						   RssTree *rssTree
						   )
{
  ObjectContainer pvs;
  if (rssTree->CollectRootPvs(pvs)) {
	Exporter *exporter = Exporter::GetExporter(filename);
	exporter->SetFilled();
	exporter->ExportGeometry(pvs);
	exporter->SetWireframe();
	exporter->ExportBox(rssTree->bbox);
	exporter->ExportViewpoint(rssTree->bbox.Center(), Vector3(1,0,0));
	delete exporter;
  }
}

bool
RssPreprocessor::ComputeVisibility()
{
  
  mSceneGraph->CollectObjects(&mObjects);
	
  long startTime = GetTime();
  
  int totalSamples = 0;

  /// Rays used for post processing and visualizations.
  RayContainer storedRays;

  AxisAlignedBox3 *box = new AxisAlignedBox3(mKdTree->GetBox());

  if (fromBoxVisibility) {
	float m = box->Min(1);
	float bsize = box->Size(1);

	float size = 0.02f;
	float s = 0.5f - size;
	float olds = Magnitude(box->Size());
	box->Enlarge(box->Size()*Vector3(-s));
	//	Vector3 translation = Vector3(-olds*0.2f, 0, 0);
	Vector3 translation = Vector3(-0.05*olds, 0, 0);
	box->SetMin(box->Min() + translation);
	box->SetMax(box->Max() + translation);

	box->SetMin(1,  m + bsize*0.1);
	box->SetMax(1,  m + bsize*0.6);

	
  } else {
		
	// sample city like heights
	float m = box->Min(1);
	float bsize = box->Size(1);
	box->SetMin(1,  m + bsize*0.2);
	box->SetMax(1,  m + bsize*0.3);
  }

  if (use2dSampling)
	box->SetMax(1, box->Min(1));
	
  if (useViewSpaceBox)
	mViewSpaceBox = box;
  else
	mViewSpaceBox = NULL;
		

  RssTree *rssTree = NULL;

  while (totalSamples < mInitialSamples) {
	int passContributingSamples = 0;
	int passSampleContributions = 0;
	int passSamples = 0;
	int index = 0;
		
	int sampleContributions;
		
	int s = Min(mSamplesPerPass, mInitialSamples);
	for (int k=0; k < s; k++) {
			
	  Vector3 viewpoint = GetViewpoint(mViewSpaceBox);
	  Vector3 direction = GetDirection(viewpoint, mViewSpaceBox);
			
	  sampleContributions = CastRay(viewpoint, direction, mVssRays);
			
			
	  //-- CORR matt: put block inside loop
	  if (sampleContributions) {
		passContributingSamples ++;
		passSampleContributions += sampleContributions;
	  }
	  passSamples++;
	  totalSamples++;
	}
    
	mPass++;
		
	int pvsSize = 0;
	float avgRayContrib = (passContributingSamples > 0) ? 
	  passSampleContributions/(float)passContributingSamples : 0;
		
	cout << "#Pass " << mPass << " : t = " << TimeDiff(startTime, GetTime())*1e-3 << "s" << endl;
	cout << "#TotalSamples=" << totalSamples/1000 
		 << "k   #SampleContributions=" << passSampleContributions << " (" 
		 << 100*passContributingSamples/(float)passSamples<<"%)" << " avgPVS="
		 << pvsSize/(float)mObjects.size() << endl 
		 << "avg ray contrib=" << avgRayContrib << endl;
		
	mStats <<
	  "#Pass\n" <<mPass<<endl<<
	  "#Time\n" << TimeDiff(startTime, GetTime())*1e-3 << endl<<
	  "#TotalSamples\n" << totalSamples<< endl<<
	  "#SampleContributions\n" << passSampleContributions << endl << 
	  "#PContributingSamples\n"<<100*passContributingSamples/(float)passSamples<<endl <<
	  "#AvgPVS\n"<< pvsSize/(float)mObjects.size() << endl <<
	  "#AvgRayContrib\n" << avgRayContrib << endl;
  }
  
  cout << "#totalPvsSize=" << mKdTree->CollectLeafPvs() << endl;
  cout << "#totalRayStackSize=" << mVssRays.size() << endl <<flush;
  

  if (numExportRays) {
	char filename[64];
	sprintf(filename, "rss-rays-initial.x3d");
	ExportRays(filename, mVssRays, numExportRays);
  }
  
  if (useViewcells)
	// construct view cells
	mViewCellsManager->Construct(mObjects, mVssRays, mViewSpaceBox);

  rssTree = new RssTree;
  // viewcells = Construct(mVssRays);
	
  if (fromBoxVisibility)
	rssTree->Construct(mVssRays, mViewSpaceBox);
  else
	rssTree->Construct(mVssRays, NULL);

  cout<<"RssTree root PVS size = "<<rssTree->GetRootPvsSize()<<endl;

  if (exportRssTree) {
	ExportRssTree("rss-tree-100.x3d", rssTree, Vector3(1,0,0));
	ExportRssTree("rss-tree-001.x3d", rssTree, Vector3(0,0,1));
	ExportRssTree("rss-tree-101.x3d", rssTree, Vector3(1,0,1));
	ExportRssTree("rss-tree-101m.x3d", rssTree, Vector3(-1,0,-1));
	ExportRssTreeLeaves(rssTree, 10);
  }

  if (exportPvs) {
	ExportPvs("rss-pvs-initial.x3d", rssTree);
  }

  // viewcells->UpdatePVS(newVssRays);
  // get viewcells as kd tree boxes
  vector<AxisAlignedBox3> kdViewcells;
  if (0) {
	vector<KdLeaf *> leaves;
	mKdTree->CollectLeaves(leaves);
	vector<KdLeaf *>::const_iterator it;
	int targetLeaves = 50;
	float prob = targetLeaves/(float)leaves.size();
	for (it = leaves.begin(); it != leaves.end(); ++it)
	  if (RandomValue(0.0f,1.0f) < prob)
		kdViewcells.push_back(mKdTree->GetBox(*it));
		
	float avgPvs = GetAvgPvsSize(rssTree, kdViewcells);
	cout<<"Initial average PVS size = "<<avgPvs<<endl;
  }

	
  int samples = 0;
  int pass = 0;
  while (1) {
	int num = mRssSamplesPerPass;
	SimpleRayContainer rays;
	VssRayContainer vssRays;
		
	if (!mUseImportanceSampling) {
	  for (int j=0; j < num; j++) {
		Vector3 viewpoint = GetViewpoint(mViewSpaceBox);
		Vector3 direction = GetDirection(viewpoint, mViewSpaceBox);
		rays.push_back(SimpleRay(viewpoint, direction));
	  }
	} else {
	  num = GenerateImportanceRays(rssTree, num, rays);
	}
		
		
	for (int i=0; i < rays.size(); i++)
	  CastRay(rays[i].mOrigin, rays[i].mDirection, vssRays);
		
	rssTree->AddRays(vssRays);
		
	if (1) {
	  int subdivided = rssTree->UpdateSubdivision();
	  cout<<"subdivided leafs = "<<subdivided<<endl;
	  cout<<"#total leaves = "<<rssTree->stat.Leaves()<<endl;

	}

	if (useViewcells) {
	  float avgPvs = GetAvgPvsSize(rssTree, kdViewcells);
	  cout<<"Average PVS size = "<<avgPvs<<endl;
	}

	if (numExportRays) {
	  char filename[64];
	  if (mUseImportanceSampling)
		sprintf(filename, "rss-rays-i%04d.x3d", pass);
	  else
		sprintf(filename, "rss-rays-%04d.x3d", pass);
	  
	  ExportRays(filename, vssRays, numExportRays);
	}

	samples+=num;

	if (useViewcells) {
	  
	  //-- prepare traversal rays for view cell intersections
	  RayContainer passRays;
	  
	  VssRayContainer::const_iterator it, it_end = vssRays.end();
	  
	  for (it = vssRays.begin(); it != it_end; ++ it)
		passRays.push_back(new Ray(*(*it)));
	  
	  int sampleContributions = 0;
	  int contributingSamples = 0;
	  
	  /// compute view cell contribution of rays
	  mViewCellsManager->ComputeSampleContributions(passRays,
													sampleContributions, 
													contributingSamples);
	  
	  //-- save rays for post processing
	  if (((int)storedRays.size() < mViewCellsManager->GetPostProcessSamples()) ||
		  ((int)storedRays.size() < mViewCellsManager->GetVisualizationSamples()))
		{
		  RayContainer::const_iterator it, it_end = passRays.end();
		  
		  for (it = passRays.begin(); it != it_end; ++ it)
			storedRays.push_back(new Ray(*(*it)));
		}
	  else
		{
		  CLEAR_CONTAINER(passRays);
		}
	  
	  float pvs = rssTree->GetAvgPvsSize();
	  cout<<"*****************************\n";
	  cout<<samples<<" avgPVS ="<<pvs<<endl;
	  cout<<"sample contributions ="<<sampleContributions<<endl;
	  cout<<"contributing sample ="<<contributingSamples<<endl;
	  cout<<"RssTree root PVS size = "<<rssTree->GetRootPvsSize()<<endl;
	  cout<<"*****************************\n";
	}
	
	if (exportPvs) {
	  char filename[64];
	  sprintf(filename, "rss-pvs-%04d.x3d", pass);
	  ExportPvs(filename, rssTree);
	}

	if (samples >= mRssSamples)
	  break;

	
	pass++;
  }
  
  if (useViewcells) {
	
	//-- post process view cells
	mViewCellsManager->PostProcess(mObjects, storedRays);
	
	//-- several visualizations and statistics
	mViewCellsManager->PrintStatistics(Debug);
	
	//-- render simulation after merge
	cout << "\nevaluating render time of final view cells ... ";
	
	const SimulationStatistics ss = mViewCellsManager->SimulateRendering();
	
	cout << " finished" << endl;
	
	cout << ss << endl;
	Debug << ss << endl;
	
	mViewCellsManager->Visualize(mObjects, storedRays);
  }
  
  delete rssTree;
  
  return true;
}

